#ifndef IOTEX_MD5_H
#define IOTEX_MD5_H

#include "build_info.h"

#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

#if !defined(IOTEX_MD5_ALT)
	// Regular implementation
	//

	/**
	 * \brief          MD5 context structure
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	typedef struct iotex_md5_context
	{
		uint32_t total[2];		  /*!< number of bytes processed  */
		uint32_t state[4];		  /*!< intermediate digest state  */
		unsigned char buffer[64]; /*!< data block being processed */
	} iotex_md5_context;

#else /* IOTEX_MD5_ALT */
	#include "md5_alt.h"
#endif /* IOTEX_MD5_ALT */

	/**
	 * \brief          Initialize MD5 context
	 *
	 * \param ctx      MD5 context to be initialized
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	void iotex_md5_init(iotex_md5_context* ctx);

	/**
	 * \brief          Clear MD5 context
	 *
	 * \param ctx      MD5 context to be cleared
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	void iotex_md5_free(iotex_md5_context* ctx);

	/**
	 * \brief          Clone (the state of) an MD5 context
	 *
	 * \param dst      The destination context
	 * \param src      The context to be cloned
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	void iotex_md5_clone(iotex_md5_context* dst, const iotex_md5_context* src);

	/**
	 * \brief          MD5 context setup
	 *
	 * \param ctx      context to be initialized
	 *
	 * \return         0 if successful
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	int iotex_md5_starts(iotex_md5_context* ctx);

	/**
	 * \brief          MD5 process buffer
	 *
	 * \param ctx      MD5 context
	 * \param input    buffer holding the data
	 * \param ilen     length of the input data
	 *
	 * \return         0 if successful
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	int iotex_md5_update(iotex_md5_context* ctx, const unsigned char* input, size_t ilen);

	/**
	 * \brief          MD5 final digest
	 *
	 * \param ctx      MD5 context
	 * \param output   MD5 checksum result
	 *
	 * \return         0 if successful
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	int iotex_md5_finish(iotex_md5_context* ctx, unsigned char output[16]);

	/**
	 * \brief          MD5 process data block (internal use only)
	 *
	 * \param ctx      MD5 context
	 * \param data     buffer holding one block of data
	 *
	 * \return         0 if successful
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	int iotex_internal_md5_process(iotex_md5_context* ctx, const unsigned char data[64]);

	/**
	 * \brief          Output = MD5( input buffer )
	 *
	 * \param input    buffer holding the data
	 * \param ilen     length of the input data
	 * \param output   MD5 checksum result
	 *
	 * \return         0 if successful
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	int iotex_md5(const unsigned char* input, size_t ilen, unsigned char output[16]);

#if defined(IOTEX_SELF_TEST)

	/**
	 * \brief          Checkup routine
	 *
	 * \return         0 if successful, or 1 if the test failed
	 *
	 * \warning        MD5 is considered a weak message digest and its use
	 *                 constitutes a security risk. We recommend considering
	 *                 stronger message digests instead.
	 *
	 */
	int iotex_md5_self_test(int verbose);

#endif /* IOTEX_SELF_TEST */

#ifdef __cplusplus
}
#endif

#endif /* iotex_md5.h */
