#ifndef IOTEX_SHA256_H
#define IOTEX_SHA256_H

#include "build_info.h"

#include <stddef.h>
#include <stdint.h>

/** SHA-256 input data was malformed. */
#define IOTEX_ERR_SHA256_BAD_INPUT_DATA -0x0074

#ifdef __cplusplus
extern "C"
{
#endif

#if !defined(IOTEX_SHA256_ALT)
	// Regular implementation
	//

	/**
	 * \brief          The SHA-256 context structure.
	 *
	 *                 The structure is used both for SHA-256 and for SHA-224
	 *                 checksum calculations. The choice between these two is
	 *                 made in the call to IOTEX_sha256_starts().
	 */
	typedef struct iotex_sha256_context
	{
	#if((IOTEX_PSA_CRYPTO_MODULE_USE) == (CRYPTO_USE_MBEDTLS))
		uint32_t total[2];		  /*!< The number of Bytes processed.  */
		uint32_t state[8];		  /*!< The intermediate digest state.  */
		unsigned char buffer[64]; /*!< The data block being processed. */
		int is224;				  /*!< Determines which function to use:
									   0: Use SHA-256, or 1: Use SHA-224. */
	#else
		void* sha256_ctx;
	#endif
	} iotex_sha256_context;

#else /* IOTEX_SHA256_ALT */
	#include "sha256_alt.h"
#endif /* IOTEX_SHA256_ALT */

	/**
	 * \brief          This function initializes a SHA-256 context.
	 *
	 * \param ctx      The SHA-256 context to initialize. This must not be \c NULL.
	 */
	void iotex_sha256_init(iotex_sha256_context* ctx);

	/**
	 * \brief          This function clears a SHA-256 context.
	 *
	 * \param ctx      The SHA-256 context to clear. This may be \c NULL, in which
	 *                 case this function returns immediately. If it is not \c NULL,
	 *                 it must point to an initialized SHA-256 context.
	 */
	void iotex_sha256_free(iotex_sha256_context* ctx);

	/**
	 * \brief          This function clones the state of a SHA-256 context.
	 *
	 * \param dst      The destination context. This must be initialized.
	 * \param src      The context to clone. This must be initialized.
	 */
	void iotex_sha256_clone(iotex_sha256_context* dst, const iotex_sha256_context* src);

	/**
	 * \brief          This function starts a SHA-224 or SHA-256 checksum
	 *                 calculation.
	 *
	 * \param ctx      The context to use. This must be initialized.
	 * \param is224    This determines which function to use. This must be
	 *                 either \c 0 for SHA-256, or \c 1 for SHA-224.
	 *
	 * \return         \c 0 on success.
	 * \return         A negative error code on failure.
	 */
	int iotex_sha256_starts(iotex_sha256_context* ctx, int is224);

	/**
	 * \brief          This function feeds an input buffer into an ongoing
	 *                 SHA-256 checksum calculation.
	 *
	 * \param ctx      The SHA-256 context. This must be initialized
	 *                 and have a hash operation started.
	 * \param input    The buffer holding the data. This must be a readable
	 *                 buffer of length \p ilen Bytes.
	 * \param ilen     The length of the input data in Bytes.
	 *
	 * \return         \c 0 on success.
	 * \return         A negative error code on failure.
	 */
	int iotex_sha256_update(iotex_sha256_context* ctx, const unsigned char* input, size_t ilen);

	/**
	 * \brief          This function finishes the SHA-256 operation, and writes
	 *                 the result to the output buffer.
	 *
	 * \param ctx      The SHA-256 context. This must be initialized
	 *                 and have a hash operation started.
	 * \param output   The SHA-224 or SHA-256 checksum result.
	 *                 This must be a writable buffer of length \c 32 bytes
	 *                 for SHA-256, \c 28 bytes for SHA-224.
	 *
	 * \return         \c 0 on success.
	 * \return         A negative error code on failure.
	 */
	int iotex_sha256_finish(iotex_sha256_context* ctx, unsigned char* output);

	/**
	 * \brief          This function processes a single data block within
	 *                 the ongoing SHA-256 computation. This function is for
	 *                 internal use only.
	 *
	 * \param ctx      The SHA-256 context. This must be initialized.
	 * \param data     The buffer holding one block of data. This must
	 *                 be a readable buffer of length \c 64 Bytes.
	 *
	 * \return         \c 0 on success.
	 * \return         A negative error code on failure.
	 */
	int iotex_internal_sha256_process(iotex_sha256_context* ctx, const unsigned char data[64]);

	/**
	 * \brief          This function calculates the SHA-224 or SHA-256
	 *                 checksum of a buffer.
	 *
	 *                 The function allocates the context, performs the
	 *                 calculation, and frees the context.
	 *
	 *                 The SHA-256 result is calculated as
	 *                 output = SHA-256(input buffer).
	 *
	 * \param input    The buffer holding the data. This must be a readable
	 *                 buffer of length \p ilen Bytes.
	 * \param ilen     The length of the input data in Bytes.
	 * \param output   The SHA-224 or SHA-256 checksum result.
	 *                 This must be a writable buffer of length \c 32 bytes
	 *                 for SHA-256, \c 28 bytes for SHA-224.
	 * \param is224    Determines which function to use. This must be
	 *                 either \c 0 for SHA-256, or \c 1 for SHA-224.
	 *
	 * \return         \c 0 on success.
	 * \return         A negative error code on failure.
	 */
	int iotex_sha256(const unsigned char* input, size_t ilen, unsigned char* output, int is224);

#ifdef __cplusplus
}
#endif

#endif /* iotex_sha256.h */
